
from .directory import (
    ADirectory,
    ADirectoryFileFinder,
    AGitDirectory,
    AGitDirectoryFileFinder)


__all__ = (
    "ADirectory",
    "ADirectoryFileFinder",
    "AGitDirectory",
    "AGitDirectoryFileFinder")




import asyncio
import contextlib
import gzip
import inspect
import os
import textwrap
from typing import (
    Any, Awaitable, Callable,
    Iterable, Iterator, List, Optional, Sized, Type, Union)

from trycast import trycast  # type:ignore

# condition needed due to https://github.com/bazelbuild/rules_python/issues/622
try:
    import orjson as json
except ImportError:
    import json  # type:ignore

from aio.core.functional import exceptions


def maybe_awaitable(result: Any) -> Awaitable:
    """Make anything awaitable.

    ```python

    async def async_fun():
        return "ASYNC FUN"

    >>> await maybe_awaitable("CABBAGE")
    'CABBAGE'
    >>> await maybe_awaitable(async_fun())
    'ASYNC FUN'

    ```
    """

    return (
        result
        if inspect.iscoroutine(result)
        else asyncio.sleep(0, result=result))


def maybe_coro(fun: Union[Callable, Awaitable]) -> Callable[..., Awaitable]:
    """Make any callable into a coroutine function.

    ```python

    def fun():
        return "FUN"

    async def async_fun():
        return "ASYNC FUN"

    >>> await maybe_coro(fun)()
    'FUN'
    >>> await maybe_coro(async_fun)()
    'ASYNC FUN'

    ```
    """

    async def async_wrapper(*args, **kwargs):
        called = fun(*args, **kwargs)
        return (
            await called
            if inspect.iscoroutinefunction(fun)
            else called)
    return async_wrapper


@contextlib.contextmanager
def nested(*contexts):
    with contextlib.ExitStack() as stack:
        yield [
            stack.enter_context(context)
            for context
            in contexts]


def junzip(data: bytes) -> Any:
    return json.loads(gzip.decompress(data))


def typed(tocast: Type, value: Any) -> Any:
    """Attempts to cast a value to a given type, TypeVar, or TypeDict.

    raises TypeError if cast value is `None`
    """

    if trycast(tocast, value) is not None:
        return value
    raise exceptions.TypeCastingError(
        "Value has wrong type or shape for Type "
        f"{tocast}: "
        f"{textwrap.shorten(str(value), width=10, placeholder='...')}")


# TODO: add async versions of the `batch` utils
def batches(items: Iterable, batch_size: int) -> Iterator[List]:
    """Yield batches of items according to batch size."""
    batch = []
    for item in items:
        batch.append(item)
        if len(batch) >= batch_size:
            yield batch
            batch = []
    if batch:
        yield batch


def batch_jobs(
        jobs: Sized,
        max_batch_size: Optional[int] = None,
        min_batch_size: Optional[int] = None) -> Iterator[List]:
    """Batch jobs between processors, optionally setting a max batch size."""
    bad_jobs_type = (
        not isinstance(jobs, Iterable)
        or isinstance(jobs, (str, bytes)))
    if bad_jobs_type:
        raise exceptions.BatchedJobsError(
            f"Wrong type for `batch_jobs` ({type(jobs)}: {jobs}")
    proc_count = os.cpu_count() or 1
    batch_count = round(len(jobs) / proc_count)
    if max_batch_size:
        batch_count = min(batch_count, max_batch_size)
    if min_batch_size:
        batch_count = max(batch_count, min_batch_size)
    return batches(typed(Iterable, jobs), batch_size=batch_count)




import functools
import time


def timing(fun):

    @functools.wraps(fun)
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        value = fun(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        print("Finished {} in {} secs".format(
            repr(fun.__name__), round(run_time, 3)))
        return value

    return wrapper




import inspect
import logging as _logging
import os
import time
from functools import cached_property, partial
from typing import Any, Callable, Optional


logger = _logging.getLogger(__name__)


class ADebugLogging:

    def __init__(self, fun: Optional[Callable] = None, *args, **kwargs):
        self.__wrapped__ = fun
        self.__doc__ = getattr(fun, '__doc__')
        self._log = kwargs.pop("log", None)
        self._format_result = kwargs.pop("format_result", None)
        self._show_cpu = kwargs.pop("show_cpu", None)

    def __call__(self, *args, **kwargs):
        if self.__wrapped__:
            if inspect.isasyncgenfunction(self.__wrapped__):
                return self.fun_async_gen(*args, **kwargs)
            elif inspect.iscoroutinefunction(self.__wrapped__):
                return self.fun_async(*args, **kwargs)
            elif inspect.isgeneratorfunction(self.__wrapped__):
                return self.fun_gen(*args, **kwargs)
            return self.fun(*args, **kwargs)
        fun: Optional[Callable] = args[0] if args else None
        self.__wrapped__ = fun
        self.__doc__ = getattr(fun, '__doc__')
        return self

    def __get__(self, instance: Any, cls=None) -> Any:
        if instance is None:
            return self
        if inspect.isasyncgenfunction(self.__wrapped__):
            return partial(self.fun_async_gen, instance)
        elif inspect.iscoroutinefunction(self.__wrapped__):
            return partial(self.fun_async, instance)
        elif inspect.isgeneratorfunction(self.__wrapped__):
            return partial(self.fun_gen, instance)
        return partial(self.fun, instance)

    @cached_property
    def name(self) -> str:
        return (
            getattr(
                self.__wrapped__, "__qualname__", None)
            or getattr(
                self.__wrapped__, "__name__", None)
            or self.__class__.__name__)

    def log(self, instance):
        if self._log:
            if isinstance(self._log, str):
                if self._log.startswith("self."):
                    return getattr(instance, self._log[5:])
                return _logging.getLogger(self._log)
            return self._log
        return logger

    def format_result(self, instance):
        if not self._format_result:
            return
        if isinstance(self._format_result, str):
            if self._format_result.startswith("self."):
                return getattr(instance, self._format_result[5:])
        return self._format_result

    def fun(self, *args, **kwargs):
        return self.log_debug_complete(
            self.log_debug_start(*args, **kwargs),
            self.__wrapped__(*args, **kwargs))

    def fun_gen(self, *args, **kwargs):
        start = self.log_debug_start(*args, **kwargs)
        count = 0
        for item in self.__wrapped__(*args, **kwargs):
            count += 1
            yield item
        self.log_debug_complete_iter(start, count)

    async def fun_async(self, *args, **kwargs):
        return self.log_debug_complete(
            self.log_debug_start(*args, **kwargs),
            await self.__wrapped__(*args, **kwargs))

    async def fun_async_gen(self, *args, **kwargs):
        start = self.log_debug_start(*args, **kwargs)
        count = 0
        async for item in self.__wrapped__(*args, **kwargs):
            count += 1
            yield item
        self.log_debug_complete_iter(start, count)

    def log_debug_start(self, instance, *args, **kwargs):
        self.log(instance).debug(
            f"{self.name} called "
            f"(args: {len(args)}, kwargs: {len(kwargs)})")
        return (instance, args, kwargs), time.perf_counter()

    def log_debug_complete(self, start, result):
        (instance, args, kwargs), start_time = start
        time_taken = time.perf_counter() - start_time
        try:
            len_info = str(len(result))
        except TypeError:
            len_info = ""
        result_info = (
            f"{type(result).__name__:5} {len_info:4} "
            f"{time_taken:6.3f}s")
        formatter = self.format_result(instance)
        if formatter:
            result_info = formatter(
                start, result, time_taken, result_info)
        self.log(instance).debug(
            f"{self.name} returns {result_info}")
        return result

    def log_debug_complete_iter(self, start, count):
        (instance, args, kwargs), start_time = start
        time_taken = time.perf_counter() - start_time
        result_info = (
            f"{type(self.__wrapped__).__name__:5} {count:4} "
            f"{time_taken:6.3f}s")
        formatter = self.format_result(instance)
        if formatter:
            result_info = formatter(
                start, count, time_taken, result_info)
        self.log(instance).debug(
            f"{self.name} "
            f"generated {result_info}")


class ATraceLogging(ADebugLogging):

    def log_debug_start(self, instance, *args, **kwargs):
        start = super().log_debug_start(instance, *args, **kwargs)
        if args:
            self.log(instance).debug(
                f"{self.name} call_args\n"
                f"   {args}")
        if kwargs:
            self.log(instance).debug(
                f"{self.name} call_kwargs\n"
                f"   {kwargs}")
        return start

    def log_debug_complete(self, start, result):
        result = super().log_debug_complete(start, result)
        (instance, args, kwargs), start_time = start
        self.log(instance).debug(
            f"{self.name} return_value\n"
            f"  {result}")
        return result


class ANullLogging(ADebugLogging):

    def log_debug_start(self, instance, *args, **kwargs):
        return (instance, args, kwargs), None

    def log_debug_complete(self, start, result):
        return result

    def log_debug_complete_iter(self, start, count):
        pass


def logging(*args, **kwargs):
    if os.environ.get("AIOTRACEDEBUG"):
        return ATraceLogging(*args, **kwargs)
    if os.environ.get("AIODEBUG"):
        return ADebugLogging(*args, **kwargs)
    return ANullLogging(*args, **kwargs)



import pathlib
from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional, Union

import aiofiles
import aiohttp

from .base import AsyncStream


class Writer(AsyncStream):
    """This wraps an async file object and provides a `stream_bytes` method to
    stream an `aiohttp.ClientResponse` to the file.

    It makes use of aiohttp's stream buffering to download chunks, and then
    writes the chunks to disk asynchronously.

    This allows large file downloads to use little or no additional
    memory while downloading with aiohttp.
    """

    async def stream_bytes(self, response: aiohttp.ClientResponse) -> None:
        """Stream chunks from an `aiohttp.ClientResponse` to an async file
        object."""
        # This is kinda aiohttp specific, we can make this more generic
        # and then adapt to aiohttp if we find the need
        async for chunk in response.content.iter_chunked(self.chunk_size):
            await self.buffer.write(chunk)


@asynccontextmanager
async def writer(
        path: Union[str, pathlib.Path],
        chunk_size: Optional[int] = None) -> AsyncIterator[Writer]:
    async with aiofiles.open(path, "wb") as f:
        yield Writer(f, chunk_size=chunk_size)



"""aio.core.functional."""

from .collections import (
    async_iterator,
    async_list,
    async_set,
    CollectionQuery,
    qdict,
    QueryDict)
from .decorators import async_property
from .generator import AwaitableGenerator
from .process import async_map
from .utils import (
    batches,
    batch_jobs,
    maybe_awaitable,
    maybe_coro,
    nested,
    typed)
from . import collections, exceptions, utils


__all__ = (
    "async_property",
    "async_iterator",
    "async_list",
    "async_map",
    "async_set",
    "AwaitableGenerator",
    "batches",
    "batch_jobs",
    "buffered",
    "CollectionQuery",
    "collections",
    "exceptions",
    "maybe_awaitable",
    "maybe_coro",
    "nested",
    "qdict",
    "QueryDict",
    "typed",
    "utils")




from typing import (
    Any, AsyncIterable, AsyncIterator, Awaitable,
    Callable, Coroutine, Dict, Generator, Iterable, Optional, Union)

from aio.core import functional


class AwaitableGenerator:
    """Wrap an asynchronous generator as optionally awaitable.

    For example, you may have a low level async function that generates
    results, but where in some cases you want to just wait for the gathered
    results.

    The 2 methods of handling results show belown are functionally equivalent,
    except that in the second example all results are generated before
    iteration begins.

    ```python

    async def some_generator():
        for x in range(0, 5):
            yield x

    async for result in AwaitableGenerator(some_generator()):
        do_something_with(result)

    for result in await AwaitableGenerator(some_generator()):
        do_something_with(result)

    ```

    By default results will be collected into a `list`, using the
    `async_list` function. You can specify a custom collector as
    an argument to the constructor.

    The generator can be called with `predicate` and `result`, which
    are optionally async callables the results of which determine
    whether to include a result and how to yield/return it respectively.

    These args are passed to the collector function when instances of
    this class are awaited.
    """

    def __init__(
            self,
            generator: AsyncIterable,
            collector: Callable[
                ...,
                Coroutine[
                    AsyncIterable,
                    None,
                    Iterable]] = None,
            iterator: Any = None,
            predicate: Optional[
                Union[
                    Callable[[Any], bool],
                    Awaitable[bool]]] = None,
            result: Optional[
                Callable[[Any], Any]] = None) -> None:
        self.generator = generator
        self.collector = collector or functional.async_list
        self.iterator = iterator or functional.async_iterator
        self.predicate = predicate
        self.result = result

    async def __aiter__(self) -> AsyncIterator:
        async for item in self.iterable:
            yield item

    def __await__(self) -> Generator:
        return self.awaitable.__await__()

    @property
    def awaitable(self) -> Awaitable:
        return self.collector(self.generator, **self.iter_kwargs)

    @property
    def iterable(self) -> AsyncIterator:
        return self.iterator(self.generator, **self.iter_kwargs)

    @property
    def iter_kwargs(self) -> Dict:
        return dict(
            predicate=self.predicate,
            result=self.result)




import os
import pathlib
from contextlib import contextmanager
from typing import Union


@contextmanager
def directory_context(path: Union[str, pathlib.Path]):
    """Sets the directory context."""
    origin = pathlib.Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)



import pathlib
from contextlib import asynccontextmanager
from typing import AsyncGenerator, AsyncIterator, Optional, Union

import aiofiles
from aiofiles.threadpool.binary import AsyncBufferedReader

from .base import AsyncStream


class Reader(AsyncStream):
    """This wraps an `AsyncBufferedReader` with a `__len__`

    This is useful if you want to a pass an `AsyncBufferedReader`
    as the body data to an `HTTP` stream, but the `HTTP` client
    wants to send a `Content-Length` header, based on the `len()`
    of the data.

    As the file's size can be gleaned from the OS `stat_size`,
    this can be set ahead of reading chunks from the file.

    This allows large file uploads to use little or no additional
    memory while uploading with aiohttp.
    """

    def __init__(self, *args, size: Optional[int] = None, **kwargs):
        self._size = size
        super().__init__(*args, **kwargs)

    async def __aiter__(self) -> AsyncGenerator[bytes, AsyncBufferedReader]:
        while True:
            chunk = await self.buffer.read(self.chunk_size)
            if not chunk:
                break
            yield chunk

    def __len__(self) -> int:
        if self.size is None:
            raise TypeError(
                f"object of type '{self.__class__.__name__}' with no 'size' "
                "cannot get len()")
        return self.size

    @property
    def size(self) -> Optional[int]:
        return self._size


@asynccontextmanager
async def reader(
        path: Union[str, pathlib.Path],
        chunk_size: Optional[int] = None) -> AsyncIterator[Reader]:
    path = pathlib.Path(path)
    async with aiofiles.open(path, "rb") as f:
        yield Reader(f, chunk_size=chunk_size, size=path.stat().st_size)



import asyncio
import inspect
import os
import types
from functools import cached_property
from typing import (
    Any, AsyncIterable, AsyncIterator, Awaitable, Callable,
    Iterable, Iterator, List,
    Optional, Union)

from aio.core.functional import async_property, AwaitableGenerator

from .exceptions import (
    ConcurrentError, ConcurrentExecutionError, ConcurrentIteratorError)


_sentinel = object()


class Concurrent:
    """This utility provides very similar functionality to
    `asyncio.as_completed` in that it runs coroutines in concurrent, yielding
    the results as they are available.

    There are a couple of differences:

    - `coros` can be any `iterables` including sync/async `generators`
    - `limit` can be supplied to specify the maximum number of concurrent tasks

    Setting `limit` to `-1` will make all tasks run in concurrent.

    The default is `number of cores + 4` to a maximum of `32`.

    For network tasks it might make sense to set the concurrency `limit` lower
    than the default, if, for example, opening many concurrent connections will
    trigger rate-limiting or soak bandwidth.

    If an error is raised while trying to iterate the provided coroutines, the
    error is wrapped in an `ConcurrentIteratorError` and is raised immediately.

    In this case, no further handling occurs, and `yield_exceptions` has no
    effect.

    Any errors raised while trying to create or run tasks are wrapped in
    `ConcurrentError`.

    Any errors raised during task execution are wrapped in
    `ConcurrentExecutionError`.

    If you specify `yield_exceptions` as `True` then the wrapped errors will be
    yielded in the results.

    If `yield_exceptions` is False (the default), then the wrapped error will
    be raised immediately.

    If you use any kind of `Generator` or `AsyncGenerator` to produce the
    awaitables, and `yield_exceptions` is `False`, in the event that an error
    occurs, it is your responsibility to `close` remaining awaitables that you
    might have created but which have not already been fired.

    This utility is mostly useful for concurrentizing io-bound (as opposed to
    cpu-bound) tasks.

    Example usage:

    ```
    import random

    from tools.base import aio

    async def task_to_run(i):
        print(f"{i} starting")
        wait = random.random() * 10
        await asyncio.sleep(wait)
        return i, wait

    async def run(coros):
        async for (i, wait) in aio.concurrent(coros, limit=3):
            print(f"{i} waited {wait}")

    def provider():
        for i in range(0, 10):
            yield task_to_run(i)

    asyncio.run(run(provider()))
    ```
    """

    def __init__(
            self,
            coros: Union[
                types.AsyncGeneratorType,
                AsyncIterable[Awaitable],
                AsyncIterator[Awaitable],
                types.GeneratorType,
                Iterator[Awaitable],
                Iterable[Awaitable]],
            yield_exceptions: Optional[bool] = False,
            limit: Optional[int] = None):
        self._coros = coros
        self._limit = limit
        self._running: List[asyncio.Task] = []
        self.yield_exceptions = yield_exceptions

    def __aiter__(self) -> AsyncIterator:
        """Start a coroutine task to process the submit queue, and return an
        async generator to deliver results back as they arrive."""
        self.submit_task = asyncio.create_task(self.submit())
        return self.output()

    @property
    def active(self) -> bool:
        """Checks whether the iterator is active, either because it hasn't
        finished submitting or because there are still tasks running."""
        return self.submitting or self.running

    @property
    def closed(self) -> bool:
        """If an unhandled error occurs, the generator is closed and no further
        processing should happen."""
        return self.closing_lock.locked()

    @cached_property
    def closing_lock(self) -> asyncio.Lock:
        """Flag to indicate whether the generator has been closed."""
        return asyncio.Lock()

    @cached_property
    def consumes_async(self) -> bool:
        """Provided coros iterable is some kind of async provider."""
        return isinstance(
            self._coros,
            (types.AsyncGeneratorType, AsyncIterator, AsyncIterable))

    @cached_property
    def consumes_generator(self) -> bool:
        """Provided coros iterable is some kind of generator."""
        return isinstance(
            self._coros,
            (types.AsyncGeneratorType, types.GeneratorType))

    @async_property
    async def coros(self) -> AsyncIterator[
            Union[ConcurrentIteratorError, Awaitable]]:
        """An async iterator of the provided coroutines."""
        coros = self.iter_coros()
        try:
            async for coro in coros:
                yield coro
        except GeneratorExit:
            # If we exit before we finish generating we land here (ie error was
            # raised). In this case we need to tell the (possibly) async
            # generating provider to also close.
            try:
                await coros.aclose()  # type:ignore
            finally:
                # Suppress errors closing the provider generator
                # This can raise a further `GeneratorExit` but it will stop
                # providing.
                return

    @property
    def default_limit(self) -> int:
        """Default is to use cpu+4 to a max of 32 coroutines."""
        # This reflects the default for asyncio's `ThreadPoolExecutor`, this is
        # a fairly arbitrary number to use, but it seems like a reasonable
        # default.
        return min(32, (os.cpu_count() or 0) + 4)

    @cached_property
    def limit(self) -> int:
        """The limit for concurrent coroutines."""
        return self._limit or self.default_limit

    @cached_property
    def nolimit(self) -> bool:
        """Flag indicating no limit to concurrency."""
        return self.limit == -1

    @cached_property
    def out(self) -> asyncio.Queue:
        """Queue of results to yield back."""
        return asyncio.Queue()

    @property
    def running(self) -> bool:
        """Flag to indicate whether any tasks are running."""
        return not self.running_queue.empty()

    @cached_property
    def running_queue(self) -> asyncio.Queue:
        """Queue which is incremented/decremented as tasks begin/end.

        This is for tracking when there are no longer any tasks running.

        A queue is used here as opposed to other synchronization primitives, as
        it allows us to get the size and emptiness.

        The queue values are `None`.
        """
        return asyncio.Queue()

    @cached_property
    def running_tasks(self) -> List[asyncio.Task]:
        """Currently running asyncio tasks."""
        return self._running

    @cached_property
    def sem(self) -> asyncio.Semaphore:
        """A sem lock to limit the number of concurrent tasks."""
        return asyncio.Semaphore(self.limit)

    @cached_property
    def submission_lock(self) -> asyncio.Lock:
        """Submission lock to indicate when submission is complete."""
        return asyncio.Lock()

    @property
    def submitting(self) -> bool:
        """Flag to indicate whether we are still submitting coroutines."""
        return self.submission_lock.locked()

    async def cancel(self) -> None:
        """Stop the submission queue, cancel running tasks, close pending
        coroutines.

        This is triggered when an unhandled error occurs and the queue
        should stop processing and bail.
        """
        # Kitchen is closed
        await self.close()

        # No more waiting
        if not self.nolimit:
            self.sem.release()

        # Cancel tasks
        await self.cancel_tasks()

        # Close pending coroutines
        await self.close_coros()

        # let the submission queue die
        await self.submit_task

    async def cancel_tasks(self) -> None:
        """Cancel any running tasks."""

        for running in self.running_tasks:
            running.cancel()
            try:
                await running
            finally:
                # ignore errors, we are dying anyway
                continue

    async def close(self) -> None:
        """Close the generator, prevent any further processing."""
        if not self.closed:
            await self.closing_lock.acquire()

    async def close_coros(self) -> None:
        """Close provided coroutines (unless the provided coros is a
        generator)"""
        if self.consumes_generator:
            # If we have a generator, dont blow/create/wait upon any more items
            return

        async for coro in self.iter_coros():
            try:
                # this could be an `aio.ConcurrentError` and not have a
                # `close` method, but as we are asking for forgiveness anyway,
                # no point in looking before we leap.
                coro.close()  # type:ignore
            finally:
                # ignore errors, we are dying anyway
                continue

    async def create_task(self, coro: Awaitable) -> None:
        """Create an asyncio task from the coroutine, and remember it."""
        task = asyncio.create_task(self.task(coro))
        self.remember_task(task)
        self.running_queue.put_nowait(None)

    async def exit_on_completion(self) -> None:
        """Send the exit signal to the output queue."""
        if not self.active and not self.closed:
            await self.out.put(_sentinel)

    def forget_task(self, task: asyncio.Task) -> None:
        """Task?

        what task?
        """
        if self.closed:
            # If we are closing, don't remove, as this has been triggered
            # by cancellation.
            return
        self.running_tasks.remove(task)

    async def iter_coros(self) -> AsyncIterator[
            Union[ConcurrentIteratorError, Awaitable]]:
        """Iterate provided coros either synchronously or asynchronously,
        yielding the awaitables asynchoronously."""
        try:
            if self.consumes_async:
                async for coro in self._coros:  # type:ignore
                    yield coro
            else:
                for coro in self._coros:  # type:ignore
                    yield coro
        except BaseException as e:
            # Catch all errors iterating (other errors are caught elsewhere)
            # If iterating raises, wrap the error and send it to `submit` and
            # and `output` to close the queues.
            yield ConcurrentIteratorError(e)

    async def on_task_complete(
            self,
            result: Any,
            decrement: Optional[bool] = True) -> None:
        """Output the result, release the sem lock, decrement the running
        count, and notify output queue if complete."""
        if self.closed:
            # Results can come back after the queue has closed as they are
            # cancelled.
            # In that case, nothing further to do.
            return

        # Give result to output
        await self.out.put(result)

        if not self.nolimit:
            # Release the sem.lock
            self.sem.release()
        if decrement:
            # Decrement the running_queue if it was incremented
            self.running_queue.get_nowait()
        # Exit if nothing left to do
        await self.exit_on_completion()

    async def output(self) -> AsyncIterator:
        """Asynchronously yield results as they become available."""
        while True:
            # Wait for some output
            result = await self.out.get()
            if result is _sentinel:
                # All done!
                await self.close()
                break
            elif self.should_error(result):
                # Raise an error and bail!
                await self.cancel()
                raise result
            yield result

    async def ready(self) -> bool:
        """Wait for the sem.lock and indicate availability in the submission
        queue."""
        if self.closed:
            return False
        if not self.nolimit:
            await self.sem.acquire()
            # We check before and after acquiring the sem.lock to see whether
            # we are `closed` as these events can be separated in
            # time/procedure.
            if self.closed:
                return False
        return True

    def remember_task(self, task: asyncio.Task) -> None:
        """Remember a scheduled asyncio task, in case it needs to be
        cancelled."""
        self.running_tasks.append(task)
        task.add_done_callback(self.forget_task)

    def should_error(self, result: Any) -> bool:
        """Check a result type and whether it should raise an error."""
        return (
            isinstance(result, ConcurrentIteratorError)
            or (isinstance(result, ConcurrentError)
                and not self.yield_exceptions))

    async def submit(self) -> None:
        """Process the iterator of coroutines as a submission queue."""
        await self.submission_lock.acquire()
        async for coro in self.coros:
            if isinstance(coro, ConcurrentIteratorError):
                # Iteration error, exit now
                await self.out.put(coro)
                break
            if not await self.ready():
                # Queue is closing, get out of here
                try:
                    # Ensure the last coro to be produced/generated is closed,
                    # as it will not be scheduled as a task, and in the case
                    # of generators it wont be closed any other way.
                    coro.close()
                finally:
                    # ignore all coro closing errors, we are dying
                    break
            # Check the supplied coro is awaitable
            try:
                self.validate_coro(coro)
            except ConcurrentError as e:
                await self.on_task_complete(e, decrement=False)
                continue
            # All good, create a task
            await self.create_task(coro)
        self.submission_lock.release()
        # If cleanup of the submission queue has taken longer than processing
        # we need to manually close
        await self.exit_on_completion()

    async def task(self, coro: Awaitable) -> None:
        """Task wrapper to catch/wrap errors and output awaited results."""
        try:
            result = await coro
        except BaseException as e:
            result = ConcurrentExecutionError(e)
        finally:
            await self.on_task_complete(result)

    def validate_coro(self, coro: Awaitable) -> None:
        """Validate that a provided coroutine is actually awaitable."""

        if not inspect.isawaitable(coro):
            raise ConcurrentError(
                f"Provided input was not a coroutine: {coro}")

        # TODO(phlax): debug this further and remove ignore.
        spent = (
            not asyncio.isfuture(coro)
            and (inspect.getcoroutinestate(coro)  # type:ignore
                 != inspect.CORO_CREATED))

        if spent:
            raise ConcurrentError(
                f"Provided coroutine has already been fired: {coro}")


def concurrent(*args, **kwargs) -> AwaitableGenerator:
    collector = kwargs.pop("collector", None)
    iterator = kwargs.pop("iterator", None)
    predicate = kwargs.pop("predicate", None)
    result = kwargs.pop("result", None)
    return AwaitableGenerator(
        Concurrent(*args, **kwargs),
        collector=collector,
        iterator=iterator,
        predicate=predicate,
        result=result)


async def inflate(
        iterable: Iterable,
        cb: Callable[[Any], Iterable[Awaitable]],
        yield_exceptions: Optional[bool] = None,
        limit: Optional[int] = None) -> AsyncIterable[Any]:
    """Inflate async data for an iterable of objects.

    The provided callback function should return an iterable of awaitables.

    For example, given a sequence of objects with an `id` property and cached
    async properties of `foo` and `bar`, this function can be used to preload
    the `foo` and `bar` properties for all the objects concurrently:

    ```python

    objects = inflate(
        inflatables,
        lambda o: (
            o.foo,
            o.bar))
    async for obj in objects:
        self.log.debug(f"Preloaded object: {obj.id}")
    ```
    """
    things = concurrent(
        (asyncio.gather(
            asyncio.sleep(0, result=thing),
            *cb(thing),
            return_exceptions=True)
         for thing
         in iterable),
        limit=limit,
        yield_exceptions=yield_exceptions)
    async for thing in things:
        if isinstance(thing[1], Exception):
            # TODO: test this further, and perhaps respect `yield_exceptions`
            raise thing[1]
        yield thing[0]




import asyncio
from concurrent import futures
from functools import cached_property
from typing import Optional

import abstracts


class IReactive(metaclass=abstracts.Interface):
    """Object that has a `loop`."""

    @property  # type: ignore
    @abstracts.interfacemethod
    def loop(self) -> asyncio.AbstractEventLoop:
        """Event loop."""
        raise NotImplementedError

    @property  # type: ignore
    @abstracts.interfacemethod
    def pool(self) -> futures.Executor:
        """Processor pool."""
        raise NotImplementedError


class AReactive(IReactive, metaclass=abstracts.Abstraction):
    _loop: Optional[asyncio.AbstractEventLoop] = None
    _pool: Optional[futures.Executor] = None

    @cached_property
    def loop(self) -> asyncio.AbstractEventLoop:
        """Event loop.

        Does not expect an existing loop to be running if it is not
        passed a loop.
        """
        return self._loop or asyncio.get_event_loop()

    @cached_property
    def pool(self) -> futures.Executor:
        return self._pool or futures.ProcessPoolExecutor()



"""aio.core.directory."""

from .context import ADirectoryContext, IDirectoryContext
from .abstract import (
    ADirectory,
    ADirectoryFileFinder,
    AGitDirectory,
    AGitDirectoryFileFinder)
from .directory import (
    Directory,
    DirectoryFileFinder,
    GitDirectory,
    GitDirectoryFileFinder)
from . import utils


__all__ = (
    "ADirectory",
    "ADirectoryContext",
    "ADirectoryFileFinder",
    "AGitDirectory",
    "AGitDirectoryFileFinder",
    "Directory",
    "DirectoryFileFinder",
    "GitDirectory",
    "GitDirectoryFileFinder",
    "IDirectoryContext",
    "utils")




from .exceptions import (
    ConcurrentError, ConcurrentExecutionError, ConcurrentIteratorError)
from .tasks import Concurrent, concurrent, inflate


__all__ = (
    "Concurrent",
    "concurrent",
    "ConcurrentError",
    "ConcurrentExecutionError",
    "ConcurrentIteratorError",
    "inflate")




from typing import Optional

from aiofiles.threadpool.binary import AsyncBufferedIOBase


class AsyncStream:
    # 16k seems to offer a good balance of performance/speed
    default_chunk_size: int = 1024 * 16

    def __init__(
            self,
            buffer: AsyncBufferedIOBase,
            chunk_size: Optional[int] = None):
        self._buffer = buffer
        self._chunk_size = chunk_size

    @property
    def buffer(self) -> AsyncBufferedIOBase:
        return self._buffer

    @property
    def chunk_size(self) -> int:
        return self._chunk_size or self.default_chunk_size




class ConcurrentError(Exception):
    """Raised when given inputs/awaitables are incorrect."""
    pass


class ConcurrentIteratorError(ConcurrentError):
    """Raised when iteration of provided awaitables fails."""
    pass


class ConcurrentExecutionError(ConcurrentError):
    """Raised when execution of a provided awaitable fails."""
    pass




import asyncio
from functools import cached_property
from typing import Any, Generator

import abstracts


class ILoader(metaclass=abstracts.Interface):
    """Maintains loading/loaded state and allows you to wait for those events
    to complete.

    Allows multiple, potentially concurrent, tasks to wait on a shared
    resource, triggering its loading only if required.

    ```python

    loader = Loader()
    data = None

    def task():
        # If the loader is loading it will wait until it completes otherwise
        # it returns immediately. The return value is the "loaded" state.
        if not await loader:
           # Loader hasnt been loaded yet
            with loader:
                data = await some_slow_task()

    asyncio.create_task(task())

    ```

    later...

    ```python

    if not await loader:
        with loader:
            data = await some_slow_task()
    ```
    """

    @abstracts.interfacemethod
    def __await__(self) -> Generator[Any, None, bool]:
        """Wait for the loader *if* it is already loading, finally returning
        "loaded" state."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def __enter__(self) -> None:
        """Start loading within a context."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def __exit__(self, *exception) -> None:
        """Complete loading exiting the context."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def loaded(self) -> bool:
        """Flag to indicate "loaded" state."""
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def loading(self) -> bool:
        """Flag to indicate "loading" state."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def complete(self) -> None:
        """Complete loading."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def start(self) -> None:
        """Start loading."""
        raise NotImplementedError

    @abstracts.interfacemethod
    async def wait(self) -> bool:
        """Wait for the loader *if* it is already loading, finally returning
        "loaded" state."""
        raise NotImplementedError


@abstracts.implementer(ILoader)
class ALoader(metaclass=abstracts.Abstraction):

    def __await__(self) -> Generator[Any, None, bool]:
        return self.wait().__await__()

    def __enter__(self) -> None:
        self.start()

    def __exit__(self, *exception) -> None:
        self.complete()

    @property
    def loaded(self) -> bool:
        return self._loaded.is_set()

    @property
    def loading(self) -> bool:
        return self._loading.is_set()

    def start(self) -> None:
        self._loading.set()

    def complete(self) -> None:

        self._loading.clear()
        self._loaded.set()

    @cached_property
    def _loaded(self) -> asyncio.Event:
        return asyncio.Event()

    @cached_property
    def _loading(self) -> asyncio.Event:
        return asyncio.Event()

    async def wait(self) -> bool:
        if self.loading:
            await self._loaded.wait()
        return self.loaded


@abstracts.implementer(ILoader)
class Loader(ALoader):
    pass




from typing import (
    Any, AsyncGenerator, AsyncIterable, Awaitable, Callable,
    Dict, Iterator, List, Mapping, Optional, Set, Tuple, Type, Union)

from aio.core.functional import exceptions
from .utils import maybe_coro, typed


async def async_iterator(
        iterable: AsyncIterable,
        predicate: Optional[
            Union[
                Callable[[Any], bool],
                Awaitable[bool]]] = None,
        result: Optional[Callable[[Any], Any]] = None) -> AsyncGenerator:
    """Iterate results of an async generator, yielding mutated results based on
    predicate."""
    result = maybe_coro(
        result
        or (lambda item: item))
    predicate_fun = (
        maybe_coro(predicate)
        if predicate
        else None)
    async for item in iterable:
        if not predicate_fun or await predicate_fun(item):
            yield await result(item)


async def async_list(
        iterable: AsyncIterable,
        predicate: Optional[
            Union[
                Callable[[Any], bool],
                Awaitable[bool]]] = None,
        result: Optional[Callable[[Any], Any]] = None) -> List:
    """Create a list from the results of an async generator."""
    results = list()
    iterator = async_iterator(iterable, predicate=predicate, result=result)
    async for item in iterator:
        results.append(item)
    return results


async def async_set(
        iterable: AsyncIterable,
        predicate: Optional[
            Union[
                Callable[[Any], bool],
                Awaitable[bool]]] = None,
        result: Optional[Callable[[Any], Any]] = None) -> Set:
    """Create a set from the results of an async generator."""
    results = set()
    iterator = async_iterator(iterable, predicate=predicate, result=result)
    async for item in iterator:
        results.add(item)
    return results


# TODO: use Mapping rather than Dict
SearchKey = Union[str, int]
CollectionQueryDict = Dict[SearchKey, str]
CollectionResultDict = Dict[str, Any]
SearchableCollection = Mapping[SearchKey, Any]
Indexable = Mapping[int, Any]


class _SearchableCollection(SearchableCollection):

    def __init__(self, collection: Mapping) -> None:
        self._collection = collection

    def __getitem__(self, k: SearchKey) -> Any:
        return self._collection.__getitem__(k)

    def __iter__(self) -> Iterator:
        return self._collection.__iter__()

    def __len__(self) -> int:
        return self._collection.__len__()


class CollectionQuery:

    def __init__(self, data: SearchableCollection) -> None:
        self.data = data

    def __call__(
            self,
            query: CollectionQueryDict) -> CollectionResultDict:
        return dict(self.iter_queries(query))

    def __getitem__(self, query: Union[str, int]) -> Any:
        return self.query(query)

    def iter_queries(
            self,
            queries: CollectionQueryDict) -> Iterator[Tuple[str, Any]]:
        for k, v in queries.items():
            yield str(k), self.query(v)

    def query(self, query: SearchKey) -> Any:
        data = self.data
        for path in self.spliterator(query):
            data = self.traverse(query, data, path)
        return data

    def traverse(self, query: SearchKey, data: Any, path: SearchKey):
        if not isinstance(path, int):
            try:
                return self.traverse_mapping(data, path)
            except KeyError as e:
                raise exceptions.CollectionQueryError(
                    f"Unable to traverse mapping {path} in {query}: {e}")
        try:
            return self.traverse_indexable(data, path)
        except IndexError as e:
            raise exceptions.CollectionQueryError(
                f"Unable to traverse index {path} in {query}: {e}")

    def spliterator(self, query: SearchKey) -> Iterator[Union[int, str]]:
        if isinstance(query, int):
            yield query
            return
        for path in query.split("/"):
            try:
                yield int(path)
            except ValueError:
                yield path

    # these could potentially use `@overload` decorator, and may not be
    # necessary. Makes mypy happy.
    def traverse_mapping(
            self,
            data: Dict[SearchKey, Any],
            key: SearchKey) -> Any:
        return data[key]

    def traverse_indexable(
            self,
            data: Indexable,
            key: int) -> Any:
        return data[key]


class QueryDict:

    def __init__(self, query: CollectionQueryDict) -> None:
        self.query = query

    def __call__(
            self,
            data: Mapping) -> CollectionResultDict:
        # TODO: optimize retrieving from same paths
        return self.query_dict(data)

    def query_dict(self, data: SearchableCollection) -> CollectionResultDict:
        return self.query_class(_SearchableCollection(data))(self.query)

    @property
    def query_class(self) -> Type[CollectionQuery]:
        return CollectionQuery


def qdict(**query) -> QueryDict:
    return QueryDict(typed(CollectionQueryDict, query))




import asyncio
import pathlib
import re
import shutil
import subprocess as subprocess
from concurrent import futures
from functools import cached_property, partial
from typing import (
    AsyncIterator, Dict, Iterable, List, Mapping, Optional,
    Pattern, Set, Tuple, Type, Union)

import abstracts

from aio.core import event, subprocess as _subprocess
from aio.core.functional import async_property, async_set, AwaitableGenerator


GREP_MIN_BATCH_SIZE = 500
GREP_MAX_BATCH_SIZE = 2000
GIT_LS_FILES_EOL_RE = r"^i/[\s]+w/(?P<eol>[a-z]*)[\s]+attr/[\s]+(?P<name>.*)"


@abstracts.implementer(_subprocess.ISubprocessHandler)
class ADirectoryFileFinder(
        _subprocess.ASubprocessHandler,
        metaclass=abstracts.Abstraction):
    """Blocking directory finder.

    Run inside a subproc, so *must* be picklable.
    """

    @classmethod
    def include_path(cls, path: str, path_matcher, exclude_matcher) -> bool:
        return bool(
            path
            and (not path_matcher
                 or path_matcher.match(path))
            and (not exclude_matcher
                 or not exclude_matcher.match(path)))

    def __init__(
            self,
            path: str,
            path_matcher: Optional[Pattern[str]] = None,
            exclude_matcher:  Optional[Pattern[str]] = None) -> None:
        _subprocess.ASubprocessHandler.__init__(self, path)
        self.path_matcher = path_matcher
        self.exclude_matcher = exclude_matcher

    def handle(self, response: subprocess.CompletedProcess) -> Set[str]:
        return set(
            path
            for path
            in self.parse_response(response)
            if self.include_path(
                path,
                self.path_matcher,
                self.exclude_matcher))

    def handle_error(
            self,
            response: subprocess.CompletedProcess) -> Dict[str, List[str]]:
        # TODO: Handle errors in directory classes
        return super().handle_error(response)

    def parse_response(
            self,
            response: subprocess.CompletedProcess) -> Iterable[str]:
        return response.stdout.split("\n")


class AGitDirectoryFileFinder(
        ADirectoryFileFinder,
        metaclass=abstracts.Abstraction):

    def __init__(
            self,
            path: str,
            path_matcher: Optional[Pattern[str]] = None,
            exclude_matcher:  Optional[Pattern[str]] = None,
            match_all_files: bool = False) -> None:
        self.match_all_files = match_all_files
        super().__init__(
            path,
            path_matcher=path_matcher,
            exclude_matcher=exclude_matcher)

    @cached_property
    def matcher(self):
        return re.compile(GIT_LS_FILES_EOL_RE)

    def parse_response(
            self,
            response: subprocess.CompletedProcess) -> Iterable[str]:
        if not self.match_all_files:
            return super().parse_response(response)
        files = []
        for line in super().parse_response(response):
            filename = self._get_file(line)
            if filename:
                files.append(filename)
        return files

    def _get_file(self, line: str) -> Optional[str]:
        eol, name = self._parse_line(line)
        if eol and eol not in ["-text", "none"]:
            return name

    def _parse_line(self, line: str) -> Tuple[Optional[str], Optional[str]]:
        matched = self.matcher.match(line)
        return (
            matched.groups()
            if matched
            else (None, None))


@abstracts.implementer(event.IExecutive)
class ADirectory(event.AExecutive, metaclass=abstracts.Abstraction):
    """A filesystem directory, with an associated fileset, and tools for
    finding files.

    The required `path` argument should be the path to an existing filesystem
    directory.

    The `exclude` and `exclude_dirs` parameters are passed to the `grep`
    command as `--exclude` and `--exclude-dir` args respectively.

    You can specify inclusive and exclusive regex matchers with `path_matcher`
    and `exclude_matcher`.

    By default, only text files (from grep/git grep pov) will be searched,
    you can override this by setting `text_only` to `False`.
    """

    def __init__(
            self,
            path: Union[str, pathlib.Path],
            exclude: Optional[Iterable[str]] = None,
            exclude_dirs: Optional[Iterable[str]] = None,
            path_matcher: Optional[Pattern[str]] = None,
            exclude_matcher: Optional[Pattern[str]] = None,
            text_only: Optional[bool] = True,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            pool: Optional[futures.Executor] = None) -> None:
        self._path = path
        self.path_matcher = path_matcher
        self.exclude_matcher = exclude_matcher
        self.text_only = text_only
        self.exclude = exclude or ()
        self.exclude_dirs = exclude_dirs or ()
        self._loop = loop
        self._pool = pool

    @cached_property
    def absolute_path(self) -> str:
        return str(self.path.resolve())

    @async_property(cache=True)
    async def files(self) -> Set[str]:
        """Set of relative file paths associated with this directory."""
        return await self.get_files()

    @cached_property
    def finder(self) -> ADirectoryFileFinder:
        return self.finder_class(
            str(self.path),
            **self.finder_kwargs)

    @property
    def finder_kwargs(self) -> Mapping:
        return dict(
            path_matcher=self.path_matcher,
            exclude_matcher=self.exclude_matcher)

    @property  # type:ignore
    @abstracts.interfacemethod
    def finder_class(self) -> Type[ADirectoryFileFinder]:
        raise NotImplementedError

    @cached_property
    def grep_args(self) -> Tuple[str, ...]:
        """Default args to pass when calling `grep`."""
        return (
            *self.grep_command_args,
            *(("-I", )
              if self.text_only
              else ()),
            *self.grep_exclusion_args)

    @property
    def grep_command_args(self) -> Tuple[str, ...]:
        """Path args for the `grep` command."""
        grep_command = shutil.which("grep")
        if grep_command:
            return grep_command, "-r"
        raise _subprocess.exceptions.OSCommandError(
            "Unable to find `grep` command")

    @property
    def grep_exclusion_args(self) -> Tuple[str, ...]:
        """Grep flags to exclude paths and directories."""
        return (
            tuple(
                f"--exclude={exclusion}"
                for exclusion
                in self.exclude)
            + tuple(
                f"--exclude-dir={directory}"
                for directory
                in self.exclude_dirs))

    @property
    def grep_max_batch_size(self) -> int:
        return GREP_MAX_BATCH_SIZE

    @property
    def grep_min_batch_size(self) -> int:
        return GREP_MIN_BATCH_SIZE

    @property
    def init_kwargs(self) -> Mapping:
        return dict(
            path=self.path,
            path_matcher=self.path_matcher,
            exclude_matcher=self.exclude_matcher,
            text_only=self.text_only,
            loop=self.loop,
            pool=self.pool)

    @cached_property
    def path(self) -> pathlib.Path:
        """Path to this directory."""
        return pathlib.Path(self._path)

    def filtered(self, **kwargs):
        init_kwargs = self.init_kwargs
        init_kwargs.update(kwargs)
        return self.__class__(**init_kwargs)

    async def get_files(self) -> Set[str]:
        return await self.grep(["-l"], "")

    def grep(
            self,
            args: Iterable,
            target: Union[str, Iterable[str]]) -> AwaitableGenerator:
        return AwaitableGenerator(
            self._grep(args, target),
            collector=async_set)

    def parse_grep_args(
            self,
            args: Iterable[str],
            target: Union[
                str,
                Iterable[str]]) -> Tuple[Tuple[str, ...], Iterable[str]]:
        """Parse `grep` args with the defaults to pass to the `grep`
        command."""
        return (
            (*self.grep_args,
             *args),
            ((target, )
             if isinstance(target, str)
             else target))

    def _batched_grep(
            self,
            grep_args: Tuple[str, ...],
            paths: Iterable[str]) -> AwaitableGenerator:
        return self.execute_in_batches(
            partial(self.finder, *grep_args),
            *paths,
            min_batch_size=self.grep_min_batch_size,
            max_batch_size=self.grep_max_batch_size)

    async def _grep(
            self,
            args: Iterable[str],
            target: Union[str, Iterable[str]]) -> AsyncIterator[str]:
        """Run `grep` in the directory."""
        if not isinstance(target, str) and not target:
            return
        batches = self._batched_grep(
            *self.parse_grep_args(args, target))
        async for batch in batches:
            for result in batch:
                yield result


class AGitDirectory(ADirectory):
    """A filesystem directory, with an associated fileset, and tools for
    finding files.

    Uses the `git` index cache for faster grepping.
    """

    @classmethod
    def find_git_deleted_files(
            cls,
            finder: AGitDirectoryFileFinder,
            git_command: str) -> Set[str]:
        return finder(
            git_command,
            "ls-files",
            "--deleted")

    @classmethod
    def find_git_files_changed_since(
            cls,
            finder: AGitDirectoryFileFinder,
            git_command: str,
            since: str) -> Set[str]:
        return finder(
            git_command,
            "diff",
            "--name-only",
            since,
            "--diff-filter=ACMR",
            "--ignore-submodules=all")

    @classmethod
    def find_git_untracked_files(
            cls,
            finder: AGitDirectoryFileFinder,
            git_command: str) -> Set[str]:
        return finder(
            git_command,
            "ls-files",
            "--others",
            "--exclude-standard",
            "--eol")

    def __init__(self, *args, **kwargs) -> None:
        self.changed = kwargs.pop("changed", None)
        self.untracked = kwargs.pop("untracked", None)
        super().__init__(*args, **kwargs)

    @async_property(cache=True)
    async def changed_files(self) -> Set[str]:
        """Files that have changed since `self.changed`, which can be the name
        of a git object (branch etc) or a commit hash."""
        return await self.execute(
            self.find_git_files_changed_since,
            self.finder,
            self.git_command,
            self.changed)

    @async_property(cache=True)
    async def deleted_files(self) -> Set[str]:
        """Files that have changed since `self.changed`, which can be the name
        of a git object (branch etc) or a commit hash."""
        return await self.execute(
            self.find_git_deleted_files,
            self.finder,
            self.git_command)

    @async_property(cache=True)
    async def files(self) -> Set[str]:
        """Files that have changed."""
        if not self.changed:
            return await self.get_files()
        # If the super `files` are going to be filtered fetch them first, and
        # only check `changed_files` if there are any matching. Otherwise do
        # the reverse - get the `changed_files` and bail if there are none.
        files = None
        if self.path_matcher or self.exclude_matcher:
            files = await self.get_files()
            if not files:
                return set()
        elif not await self.changed_files:
            return set()
        return (
            (await self.get_files()
             if files is None
             else files)
            & await self.changed_files)

    @property
    def finder_kwargs(self) -> Mapping:
        return dict(
            **super().finder_kwargs,
            match_all_files=self.untracked)

    @property
    def git_command(self) -> str:
        """Path to the `git` command."""
        git_command = shutil.which("git")
        if git_command:
            return git_command
        raise _subprocess.exceptions.OSCommandError(
            "Unable to find the `git` command")

    @property
    def grep_command_args(self) -> Tuple[str, ...]:
        return self.git_command, "grep", "--cached"

    @property
    def init_kwargs(self) -> Mapping:
        return dict(**super().init_kwargs, untracked=self.untracked)

    @async_property
    async def untracked_files(self):
        return await self.execute(
            self.find_git_untracked_files,
            self.finder,
            self.git_command)

    async def get_files(self) -> Set[str]:
        return (
            await super().get_files() - await self.deleted_files
            if not self.untracked
            else await self.untracked_files)




from . import (
    dev,
    directory,
    event,
    functional,
    stream,
    subprocess,
    tasks)


__all__ = (
    "dev",
    "directory",
    "event",
    "functional",
    "stream",
    "subprocess",
    "tasks")



"""aio.core.subprocess."""

from .async_subprocess import AsyncSubprocess, run, parallel
from .handler import ASubprocessHandler, ISubprocessHandler
from . import exceptions


__all__ = (
    "exceptions",
    "run",
    "parallel",
    "ASubprocessHandler",
    "AsyncSubprocess",
    "ISubprocessHandler")



"""aio.core.event."""

from .loader import ALoader, ILoader, Loader
from .reactive import AReactive, IReactive
from .executive import AExecutive, IExecutive


__all__ = (
    "AExecutive",
    "ALoader",
    "AReactive",
    "Loader",
    "IExecutive",
    "ILoader",
    "IReactive")




import abc
import logging
import subprocess
from functools import cached_property
from typing import Any, Mapping, Sequence, Tuple

import abstracts

from aio.core import directory
from aio.core.dev import debug


class ISubprocessHandler(
        directory.IDirectoryContext,
        metaclass=abstracts.Interface):
    """Protocol for handling subprocess calls."""

    @property  # type:ignore
    @abstracts.interfacemethod
    def args(self) -> Sequence[str]:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def encoding(self) -> str:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def kwargs(self) -> Mapping:
        raise NotImplementedError

    @abstracts.interfacemethod
    def handle(self, response: subprocess.CompletedProcess) -> Any:
        """Handle a successful response."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def handle_error(self, response: subprocess.CompletedProcess) -> Any:
        """Handle a failing response."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def handle_response(self, response: subprocess.CompletedProcess) -> Any:
        """Handle a response."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def has_failed(self, response: subprocess.CompletedProcess) -> bool:
        raise NotImplementedError

    @abstracts.interfacemethod
    def run(self,
            *args: str,
            **kwargs) -> Any:
        """Run the subprocess, returning handled results."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def run_subprocess(
            self,
            *args: str,
            **kwargs) -> subprocess.CompletedProcess:
        """Run the subprocess."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def subprocess_args(self, *args: str) -> Tuple[Sequence[str], ...]:
        """Derive the subprocess args, from init args and supplied args."""
        raise NotImplementedError

    @abstracts.interfacemethod
    def subprocess_kwargs(self, **kwargs):
        """Derive the subprocess kwargs, from init kwargs and supplied
        kwargs."""
        raise NotImplementedError


@abstracts.implementer(ISubprocessHandler)
class ASubprocessHandler(
        directory.ADirectoryContext,
        metaclass=abstracts.Abstraction):
    _args: Sequence[str]
    _kwargs: Mapping
    _encoding: str

    def __init__(
            self,
            path: str,
            *args: str,
            encoding: str = "utf-8",
            **kwargs) -> None:
        directory.ADirectoryContext.__init__(self, path)
        self._encoding = encoding
        self._args = args
        self._kwargs = kwargs

    def __call__(self, *args: str) -> Any:
        return self.run(*args)

    def __str__(self) -> str:
        return f"{self.__class__.__module__}.{self.__class__.__name__}"

    @property
    def args(self) -> Sequence[str]:
        return (
            self._args
            if self._args
            else ())

    @property
    def encoding(self) -> str:
        return self._encoding

    @property
    def kwargs(self) -> Mapping:
        return dict(
            cwd=self.path,
            capture_output=True,
            encoding=self.encoding)

    @cached_property
    def log(self) -> logging.Logger:
        """Logger to use - derived from implementer name."""
        return logging.getLogger(str(self))

    @abc.abstractmethod
    def handle(
            self,
            response: subprocess.CompletedProcess) -> Any:
        return dict(
            RESPONSE=[
                response.returncode,
                response.stdout,
                response.stderr])

    @abc.abstractmethod
    def handle_error(
            self,
            response: subprocess.CompletedProcess) -> Any:
        return dict(
            ERROR=[
                response.returncode,
                response.stdout,
                response.stderr])

    def handle_response(
            self,
            response: subprocess.CompletedProcess) -> Any:
        return (
            self.handle_error(response)
            if self.has_failed(response)
            else self.handle(response))

    def has_failed(self, response: subprocess.CompletedProcess) -> bool:
        return bool(response.returncode)

    @debug.logging(
        log="self.log",
        show_cpu=True)
    def run(self, *args, **kwargs) -> Any:
        """Run the subprocess and handle the results."""
        return self.handle_response(
            self.run_subprocess(
                *self.subprocess_args(*args),
                **self.subprocess_kwargs(**kwargs)))

    def run_subprocess(
            self,
            *args,
            **kwargs) -> subprocess.CompletedProcess:
        return subprocess.run(*args, **kwargs)

    def subprocess_args(self, *args) -> Tuple[Sequence[str], ...]:
        return ((*self.args, *args), )

    def subprocess_kwargs(self, **kwargs) -> Mapping:
        return {**self.kwargs, **kwargs}




from typing import Type

import abstracts

from aio.core import directory


@abstracts.implementer(directory.ADirectory)
class Directory:

    @property
    def finder_class(self) -> Type[directory.ADirectoryFileFinder]:
        return DirectoryFileFinder


@abstracts.implementer(directory.ADirectoryFileFinder)
class DirectoryFileFinder:
    pass


@abstracts.implementer(directory.AGitDirectoryFileFinder)
class GitDirectoryFileFinder:
    pass


@abstracts.implementer(directory.AGitDirectory)
class GitDirectory:

    @property
    def finder_class(self) -> Type[directory.AGitDirectoryFileFinder]:
        return GitDirectoryFileFinder




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: aio.core:package

from setuptools import setup

setup(**{
    'author': 'Ryan Northey',
    'author_email': 'ryan@synca.io',
    'description': 'A collection of functional utils for asyncio',
    'install_requires': (
        'abstracts>=0.0.12',
        'aiofiles',
        'aiohttp>=3.8.1',
        'orjson',
        'trycast>=0.7.3',
        'types-orjson',
    ),
    'license': 'Apache Software License 2.0',
    'long_description': """
aio.core
========

Utils for asyncio.
""",
    'maintainer': 'Ryan Northey',
    'maintainer_email': 'ryan@synca.io',
    'name': 'aio.core',
    'namespace_packages': (
    ),
    'package_data': {
        'aio.core': (
            'py.typed',
        ),
    },
    'packages': (
        'aio.core',
        'aio.core.dev',
        'aio.core.directory',
        'aio.core.directory.abstract',
        'aio.core.event',
        'aio.core.functional',
        'aio.core.log',
        'aio.core.stream',
        'aio.core.subprocess',
        'aio.core.tasks',
    ),
    'url': 'https://github.com/envoyproxy/pytooling/tree/main/aio.core',
    'version': '0.8.6',
})




from functools import partial
from typing import Any, Callable, Optional

import abstracts

from aio.core import event, functional, tasks
from aio.core.dev import debug


# TODO: split `IReactive.pool` to here
class IExecutive(event.IReactive, metaclass=abstracts.Interface):
    """Object that executes commands in a process pool."""

    @abstracts.interfacemethod
    async def execute(
            self,
            executable: Callable,
            *args,
            **kwargs) -> Any:
        """Execute a command in a process pool."""
        raise NotImplementedError

    @abstracts.interfacemethod
    async def execute_in_batches(
            self,
            executable: Callable,
            *args,
            concurrency: Optional[int] = None,
            min_batch_size: Optional[int] = None,
            max_batch_size: Optional[int] = None,
            **kwargs) -> functional.AwaitableGenerator:
        """Execute a command in a process pool."""
        raise NotImplementedError


@abstracts.implementer((event.AReactive, IExecutive))
class AExecutive(metaclass=abstracts.Abstraction):

    @debug.logging(
        log=__name__,
        format_result="self._debug_execute")
    async def execute(
            self,
            executable: Callable,
            *args,
            **kwargs) -> Any:
        return await self.loop.run_in_executor(
            self.pool,
            *(executable, *args),
            **kwargs)

    def execute_in_batches(
            self,
            executable: Callable,
            *args,
            concurrency: Optional[int] = None,
            min_batch_size: Optional[int] = None,
            max_batch_size: Optional[int] = None,
            **kwargs) -> functional.AwaitableGenerator:
        return tasks.concurrent(
            (self.execute(
                executable,
                *batch,
                **kwargs)
             for batch
             in functional.batch_jobs(
                 args,
                 min_batch_size=min_batch_size,
                 max_batch_size=max_batch_size)),
            limit=concurrency)

    def _debug_execute(self, start, result, time_taken, result_info):
        (instance, (executable, *args), kwargs), start_time = start
        pool_name = (
            "\N{forking}"
            if self.pool.__class__.__name__ == "ProcessPoolExecutor"
            else "\N{nonforking}")
        pool_info = f"{pool_name}:{hex(id(self.pool))}"
        if type(executable) == partial:
            executable = executable.func
        name = getattr(
            executable, "__qualname__",
            executable.__class__.__qualname__)
        return (
            f"{pool_info} {result_info}: "
            f"{executable.__module__}{name}")



#
# Decorators
#

import inspect
import logging
from functools import cached_property
from typing import Any, Callable, Dict, Optional

from aio.core import event
from aio.core.dev import debug


logger = logging.getLogger(__name__)


class NoCache(Exception):
    pass


class async_property:  # noqa: N801
    name = None
    cache_name = "__async_prop_cache__"

    @classmethod
    def is_cached(cls, item: object, name: str) -> bool:
        return (
            hasattr(item, cls.cache_name)
            and name in getattr(item, cls.cache_name))

    # If the decorator is called with `kwargs` then `fun` is `None`
    # and instead `__call__` is triggered with `fun`
    def __init__(self, fun: Optional[Callable] = None, cache: bool = False):
        self.cache = cache
        self._fun = fun
        self.name = getattr(fun, "__name__", None)
        self.__doc__ = getattr(fun, '__doc__')
        if fun and hasattr(fun, "__isabstractmethod__"):
            self.__isabstractmethod__ = fun.__isabstractmethod__  # type:ignore
        if fun and hasattr(fun, "__isinterfacemethod__"):
            self.__isinterfacemethod__ = (
                fun.__isinterfacemethod__)  # type:ignore

    def __call__(self, fun: Callable) -> 'async_property':
        self._fun = fun
        self.name = self.name or fun.__name__
        self.__doc__ = getattr(fun, '__doc__')
        if hasattr(fun, "__isabstractmethod__"):
            self.__isabstractmethod__ = fun.__isabstractmethod__  # type:ignore
        if hasattr(fun, "__isinterfacemethod__"):
            self.__isinterfacemethod__ = (
                fun.__isinterfacemethod__)  # type:ignore
        return self

    def __get__(self, instance: Any, cls=None) -> Any:
        if instance is None:
            return self
        if inspect.isasyncgenfunction(self._fun):
            return self.async_iter_result(instance)
        return self.async_result(instance)

    @cached_property
    def loaders(self) -> Dict[str, event.Loader]:
        return {}

    def fun(self, *args, **kwargs):
        if self._fun:
            return self._fun(*args, **kwargs)

    def get_prop_cache(self, instance: Any) -> dict:
        return getattr(instance, self.cache_name, {})

    # An async wrapper function to return the result
    # This is returned when the prop is called if the wrapped
    # method is an async generator
    async def async_iter_result(self, instance: Any):
        # retrieve the value from cache if available
        try:
            result = self.get_cached_prop(instance)
        except (NoCache, KeyError):
            result = None

        if result is None:
            result = self.set_prop_cache(instance, self.fun(instance))

        async for item in result:
            yield item

    # An async wrapper function to return the result
    # This is returned when the prop is called
    async def async_result(self, instance: Any) -> Any:
        # retrieve the value from cache if available
        try:
            return self.get_cached_prop(instance)
        except (NoCache, KeyError):
            pass
        if self.cache:
            return await self.load_cooperatively(instance)
        return await self.load(instance)

    def get_cached_prop(self, instance: Any) -> Any:
        if not self.cache:
            raise NoCache
        return self.get_prop_cache(instance)[self.name]

    def get_loader(self, instance: Any) -> Optional[event.Loader]:
        try:
            hash(instance)
        except TypeError:
            return None
        if instance not in self.loaders:
            self.loaders[instance] = event.Loader()
        return self.loaders[instance]

    @debug.logging(
        log=__name__,
        format_result="self._debug_prop")
    async def load(self, instance: Any) -> Any:
        # derive the result, set the cache if required, and return the result
        return self.set_prop_cache(instance, await self.fun(instance))

    async def load_cooperatively(self, instance: Any) -> Any:
        loader = self.get_loader(instance)
        if not loader:
            # Unhashable type, just load
            return await self.load(instance)
        if not await loader:
            with loader:
                return await self.load(instance)
        return self.get_cached_prop(instance)

    def set_prop_cache(self, instance: Any, result: Any) -> Any:
        if not self.cache:
            return result
        cache = self.get_prop_cache(instance)
        cache[self.name] = result
        setattr(instance, self.cache_name, cache)
        return result

    def _debug_prop(self, start, result, time_taken, result_info):
        (instance, args, kwargs), start_time = start
        cache = (
            "\N{viewdata square}"
            if self.cache
            else "~")
        return f"{cache} {result_info}: {self._repr(instance)}"

    def _repr(self, instance: Any) -> str:
        if self._fun is None:
            return "async_property"
        return (
            f"<{instance.__class__.__module__}"
            f".{self._fun.__qualname__} {hex(id(instance))}>")




import asyncio
import atexit
import logging
import logging.handlers
from functools import cached_property
from queue import SimpleQueue
from typing import List, Type


class QueueHandler(logging.handlers.QueueHandler):

    def emit(self, record: logging.LogRecord) -> None:
        try:
            self.enqueue(record)
        except asyncio.CancelledError:
            raise
        except Exception:
            self.handleError(record)


class QueueLogger:
    """Wraps a `logging.Logger` with a listening queue.

    Calling the `start` method returns the original `Logger`, with the
    listener started.

    If you set `stop_on_exit` to `False`, you must `stop` the listener
    yourself to ensure the logging queue is cleared.
    """

    def __init__(
            self,
            logger: logging.Logger,
            stop_on_exit: bool = True,
            respect_handler_level: bool = True) -> None:
        self._logger = logger
        self.respect_handler_level = respect_handler_level
        self.stop_on_exit = stop_on_exit

    @cached_property
    def handler(self) -> QueueHandler:
        return self.handler_class(self.queue)

    @property
    def handler_class(self) -> Type[QueueHandler]:
        return QueueHandler

    @cached_property
    def handlers(self) -> List[logging.Handler]:
        """Logging handlers removed from the original logger."""
        handlers: List[logging.Handler] = []
        for h in self.logger.handlers[:]:
            if h is not self.handler:
                self.logger.removeHandler(h)
                handlers.append(h)
        return handlers

    @property
    def listener_class(self) -> Type[logging.handlers.QueueListener]:
        return logging.handlers.QueueListener

    @cached_property
    def queue(self) -> SimpleQueue:
        return self.queue_class()

    @property
    def queue_class(self) -> Type[SimpleQueue]:
        return SimpleQueue

    @cached_property
    def listener(self) -> logging.handlers.QueueListener:
        return self.listener_class(
            self.queue,
            *self.handlers,
            respect_handler_level=self.respect_handler_level)

    @cached_property
    def logger(self) -> logging.Logger:
        """Wrapped `Logger` with `handler` added."""
        self._logger.addHandler(self.handler)
        return self._logger

    def start(self) -> logging.Logger:
        self.listener.start()
        if self.stop_on_exit:
            atexit.register(self.listener.stop)
        return self.logger




from concurrent import futures
from typing import AsyncIterator, Callable, Iterable


async def async_map(
        fun: Callable,
        iterable: Iterable,
        fork: bool = False) -> AsyncIterator:
    """Asynchronously map synchronous function with an iterable, yielding
    results as available.

    By default a `ThreadPoolExecutor` will be used, setting `fork` to
    `True` will instead use a `ProcessPoolExecutor` to fork the function
    to separate processes.
    """
    executor = (
        futures.ProcessPoolExecutor
        if fork
        else futures.ThreadPoolExecutor)

    with executor() as pool:
        result_futures = list(map(lambda x: pool.submit(fun, x), iterable))
        for future in futures.as_completed(result_futures):
            yield future.result()



import asyncio
import subprocess
from concurrent.futures import (
    Executor, ProcessPoolExecutor, ThreadPoolExecutor)
from functools import partial
from typing import AsyncGenerator, Iterable, Optional


from aio.core import functional


class AsyncSubprocess:

    @classmethod
    async def parallel(
            cls,
            commands: Iterable[Iterable[str]],
            fork: bool = True,
            **kwargs) -> AsyncGenerator[
                subprocess.CompletedProcess,
                Iterable[Iterable[str]]]:
        """Run external subprocesses in parallel.

        Yields `subprocess.CompletedProcess` results as they are completed.

        Example usage:

        ```
        import asyncio

        from aio.core.subprocess import parallel

        async def run_system_commands(commands):
            async for result in parallel(commands, capture_output=True):
                print(result.returncode)
                print(result.stdout)
                print(result.stderr)

        asyncio.run(run_system_commands(["whoami"] for i in range(0, 5)))
        ```
        """
        # Using a `ProcessPoolExecutor` or `ThreadPoolExecutor` here is
        # somewhat arbitrary as subproc will spawn a new process regardless.
        # Either way - using a custom executor of either type gives
        # considerable speedup, most likely due to the number of workers
        # allocated. In my testing, `ProcessPoolExecutor` gave a very small
        # speedup over a large number of tasks, despite any additional overhead
        # of creating the executor. Without `max_workers` set
        # `ProcessPoolExecutor` defaults to the number of cpus on the machine.
        executor = (
            ProcessPoolExecutor
            if fork
            else ThreadPoolExecutor)
        with executor() as pool:
            futures = asyncio.as_completed(
                tuple(
                    asyncio.ensure_future(
                        cls.run(command, executor=pool, **kwargs))
                    for command in commands))
            for result in futures:
                yield await result

    @classmethod
    async def run(
            cls,
            *args,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            executor: Optional[Executor] = None,
            **kwargs) -> subprocess.CompletedProcess:
        """This is an asyncio wrapper for `subprocess.run`

        It can be used in a similar way to `subprocess.run` but its
        non-blocking to the main thread.

        Example usage:

        ```
        import asyncio

        from aio.core.subprocess import run

        async def run_system_command():
            result = await run(["whoami"], capture_output=True)
            print(result.returncode)
            print(result.stdout)
            print(result.stderr)

        asyncio.run(run_system_command())

        ```

        By default it will spawn the process using the main event loop, and
        that loop's default (`ThreadPool`) executor.

        You can provide the loop and/or the executor to change this behaviour.
        """
        loop = loop or asyncio.get_running_loop()
        return await loop.run_in_executor(
            executor, partial(subprocess.run, *args, **kwargs))


def parallel(*args, **kwargs) -> "functional.AwaitableGenerator":
    collector = kwargs.pop("collector", None)
    iterator = kwargs.pop("iterator", None)
    predicate = kwargs.pop("predicate", None)
    result = kwargs.pop("result", None)
    return functional.AwaitableGenerator(
        AsyncSubprocess.parallel(*args, **kwargs),
        collector=collector,
        iterator=iterator,
        predicate=predicate,
        result=result)


run = AsyncSubprocess.run


__all__ = (
    "AsyncSubprocess",
    "parallel",
    "run")




import os
import pathlib
from functools import cached_property
from typing import Optional, Union

import abstracts

from . import utils


class IDirectoryContext(metaclass=abstracts.Interface):
    """Directory context has `path` and an `in_directory` contextmanager."""

    @property  # type:ignore
    @abstracts.interfacemethod
    def path(self) -> pathlib.Path:
        raise NotImplementedError

    @property  # type:ignore
    @abstracts.interfacemethod
    def in_directory(self):
        """Context manager that `chdir`s to `self.path`."""
        raise NotImplementedError


@abstracts.implementer(IDirectoryContext)
class ADirectoryContext(metaclass=abstracts.Abstraction):
    _path: Optional[Union[str, os.PathLike]] = None

    def __init__(
            self,
            path: Union[str, os.PathLike]) -> None:
        self._path = path

    @cached_property
    def path(self) -> pathlib.Path:
        return pathlib.Path(self._path or ".")

    @property
    def in_directory(self):
        return utils.directory_context(self.path)
